import { Database, open } from 'sqlite'
import sqlite3 from 'sqlite3'
import { ErrStorage, Storage, StorageCursor, StorageSession } from '../storage'
import { ErrEntityHasNoUniqKeyValue } from '../entity'

export type SQLiteCriteria = Record<string, Array<string | number>>

function parseCriteria(criteria?: SQLiteCriteria): { where: string | null; params: Array<string | number> } {
    if (criteria) {
        const conditions = Object.entries(criteria)
        const params = []
        const where = []
        if (conditions.length) {
            for (const condition of conditions) {
                where.push(condition[0].trim())
                params.push(...condition[1])
            }
            return { where: where.join(' AND '), params }
        }
    }
    return { where: null, params: [] }
}

function quoteStr(val: string | number | unknown) {
    if (typeof val === 'string') {
        return `'${val.replace(/'/g, `"`)}'`
    }
    return val
}

function transformDate(val: Date | unknown) {
    return val instanceof Date ? `'${val.getTime()}'` : val
}

export interface StorageSQLiteParams {
    dsn: string
}

export class StorageSQLite extends Storage {
    private _db: Database | null = null

    constructor(params: StorageSQLiteParams) {
        super(params.dsn)
    }

    async connect(): Promise<Database> {
        try {
            sqlite3.verbose()
            this._db = await open({
                filename: this._dsn,
                driver: sqlite3.Database,
            })
            return this._db
        } catch (e) {
            throw new ErrStorage(`error on connect to db [${this._dsn}]`)
        }
    }

    async count(name: string, criteria?: SQLiteCriteria): Promise<number> {
        if (!this._db) await this.connect()
        const query = [`SELECT COUNT(*) as cnt FROM ${name}`]
        const { where, params } = parseCriteria(criteria)
        if (where) {
            query.push(`WHERE ${where}`)
        }
        const result = await this._db?.get(query.join(' '), params)
        return parseInt(result.cnt, 10)
    }

    async find(name: string, criteria: SQLiteCriteria): Promise<SQLiteCursor> {
        if (!this._db) this._db = await this.connect()
        return new SQLiteCursor(this._db, name, criteria)
    }

    /**
     * Вернет тру, даже если записи не было
     * @param name
     * @param uniqKey
     * @param uniqValue
     */
    async remove(name: string, uniqKey: string, uniqValue: string): Promise<boolean> {
        if (!this._db) this._db = await this.connect()
        const sql = `DELETE FROM ${name} WHERE ${uniqKey}=${quoteStr(uniqValue)}`
        const result = await this._db.run(sql)
        return !!(result && result.changes)
    }

    async save(name: string, uniqKey: string, data: Record<string, unknown>): Promise<string> {
        if (!this._db) this._db = await this.connect()
        const keys = []
        const values = []
        const assigns = []
        for (const [k, v] of Object.entries(data)) {
            if (v === undefined) continue

            keys.push(k)

            let vp
            if (v === null) {
                vp = 'NULL'
            } else {
                vp = quoteStr(v)
                vp = transformDate(vp)
            }
            values.push(vp)
            assigns.push(`${k}=${vp}`)
        }
        const id = data[uniqKey]
        if (id !== null && id !== undefined) {
            const sql = `INSERT INTO ${name}(${keys.join(', ')})
					 VALUES (${values.join(', ')})
					 ON CONFLICT(${uniqKey}) DO UPDATE SET ${assigns.join(', ')}
					 WHERE ${uniqKey} = ${quoteStr(id)}`
            try {
                const result = await this._db.run(sql)
                return result.lastID + ''
            } catch (e) {
                throw new ErrStorage(`can not save data to ${name} by sql:\n"${sql}"\nwith error: ${e}`)
            }
        } else {
            throw new ErrEntityHasNoUniqKeyValue()
        }
    }

    createSession(): StorageSession {
        return new SQLiteSession(this._db)
    }

    get db(): Database | null {
        return this._db
    }
}

/**
 *
 */
export class SQLiteCursor implements StorageCursor {
    private _limit = 0
    private _offset = 0
    private _order: Record<string, string> = {}
    constructor(private _db: Database, private _tableName: string, private _criteria?: SQLiteCriteria) {}

    limit(number: number): this {
        this._limit = number
        return this
    }

    skip(offset: number): this {
        this._offset = offset
        return this
    }

    sort(parameters: Record<string, string>): this {
        this._order = parameters
        return this
    }

    _buildSQL(): { query: string[]; params: (string | number)[] } {
        const query = [`SELECT ROWID as '_id', * FROM ${this._tableName}`]
        const { where, params } = parseCriteria(this._criteria)
        if (where) {
            query.push(`WHERE ${where}`)
        }

        const orders = Object.entries(this._order)
        if (orders.length) {
            const orderStr = orders.map(e => `${e[0]} ${e[1]}`).join(', ')
            query.push(`ORDER BY ${orderStr}`)
        }

        if (this._limit) query.push(`LIMIT ${this._limit}`)
        if (this._offset) query.push(`OFFSET ${this._offset}`)

        return { query, params }
    }

    async toArray(): Promise<Record<string, any>[]> {
        const { query, params } = this._buildSQL()
        return this._db.all(query.join(' '), params.length ? params : [])
    }
}

/**
 *
 */
export class SQLiteSession extends StorageSession {
    constructor(private _db: Database | null) {
        super()
    }

    async start(): Promise<void> {
        await this._db?.run('BEGIN TRANSACTION')
        return
    }

    async commit(fn: () => void): Promise<void> {
        try {
            await fn()
        } catch (e) {
            await this._db?.run('ROLLBACK')
            throw e
        }

        await this._db?.run('COMMIT')
        return
    }
}
